""" Classes to handle configuration of a PubTal site.

	Copyright (c) 2003 Colin Stewart (http://www.owlfish.com/)
	All rights reserved.
		
	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:
	1. Redistributions of source code must retain the above copyright
	   notice, this list of conditions and the following disclaimer.
	2. Redistributions in binary form must reproduce the above copyright
	   notice, this list of conditions and the following disclaimer in the
	   documentation and/or other materials provided with the distribution.
	3. The name of the author may not be used to endorse or promote products
	   derived from this software without specific prior written permission.
	
	THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
	IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
	
	If you make any bug fixes or feature enhancements please let me know!

"""

try:
	import logging
except:
	import InfoLogging as logging
	
import pubtal
import ConfigurationParser, BuiltInPlugins, EncodingCapabilities
import os, os.path, stat, re, sys
	
class SiteConfig:
	def __init__ (self, configFile):
		self.log = logging.getLogger ("PubTal.SiteConfig")
		self.contentConfig = ContentConfig (self)
		self.templateConfig = TemplateConfig (self)
		self.encodingCapabilities = EncodingCapabilities.EncodingCapabilities()
		# List of regular expressions that match files to be ignored
		self.ignoreFilters = []
		# Directories where plugins may reside
		self.extraPluginDirs = []
		self.plugins = []
		self.baseDir = os.path.abspath (os.path.split (configFile)[0])
		
		# Defaults
		self.setDefaultCharacterSet ('ISO-8859-15')
		self.contentDir = os.path.join (self.baseDir, 'content')
		self.destDir = os.path.join (self.baseDir, 'dest')
		self.templateDir = os.path.join (self.baseDir, 'template')
		
		self.readConfig(configFile)
		
		# Load plugins
		self.log.info ("Loading plugins...")
		systemPlugins = self.getPluginModules(os.path.join (pubtal.__path__[0],'plugins'))
		self.plugins.extend (systemPlugins)
		self.log.debug ("Loading extra plugins.")
		for extraDir in self.extraPluginDirs:
			self.log.debug ("Looking for plugins in %s" % extraDir)
			morePlugins = self.getPluginModules (extraDir)
			self.plugins.extend (morePlugins)
		
		# Add in the built in supported content
		self.plugins.append (BuiltInPlugins)
		self.supportedContent = {}
		for plugin in self.plugins:
			for pluginInfo in plugin.getPluginInfo():
				pluginType = pluginInfo.get ('functionality', None)
				if (pluginType == None):
					sef.log.warn ("Plugin did not include 'functionality' key - skipping.")
				elif (pluginType == "content"):
					contentType = pluginInfo ['content-type']
					klass = pluginInfo ['class']
					self.supportedContent [contentType] = klass
					if (pluginInfo.has_key ('file-type')):
						contentTypeConfig = PageConfigItem()
						contentTypeConfig.setOption ('content-type', contentType)
						self.log.debug ("Adding plugin file-type config for %s" % pluginInfo ['file-type'])
						self.contentConfig.addFileType (pluginInfo ['file-type'], contentTypeConfig)
				else:
					self.log.warn ("Plugin offer functionality %s which is not understood." % pluginType)
		
	def getContentDir (self):
		return self.contentDir
		
	def getDestinationDir (self):
		return self.destDir
		
	def getTemplateDir (self):
		return self.templateDir
		
	def getIgnoreFilters (self):
		return self.ignoreFilters
		
	def getDefaultCharacterSet (self):
		return self.characterSet
		
	def setDefaultCharacterSet (self, charSet):
		self.characterSet = charSet
		
	def getSupportedContent (self):
		return self.supportedContent
		
	def getEncodingCapabilities (self):
		return self.encodingCapabilities
		
	def getContentConfig (self):
		return self.contentConfig
	
	def getTemplateConfig (self):
		return self.templateConfig
		
	def readConfig (self, configFile):
		parser = ConfigurationParser.ConfigurationParser()
		parser.addHandler ('SiteConfig', self)
		parser.addHandler ('Content', self.contentConfig)
		parser.addHandler ('Template', self.templateConfig)
		
		self.currentDirective = None
		confFile = open (configFile, 'r')
		parser.parse(confFile)
		confFile.close()
		
		self.contentConfig.configFinished()
		self.templateConfig.configFinished()
		
	def startDirective (self, directive, options):
		self.currentDirective = directive
		
	def endDirective (self, directive):
		self.currentDirective = None
		
	def option (self, line):
		if (self.currentDirective == 'SITECONFIG'):
			if (line.lower().startswith ('content-dir')):
				self.contentDir = os.path.join (self.baseDir, line [line.find (' ')+1:])
			elif (line.lower().startswith ('template-dir')):
				self.templateDir = os.path.join (self.baseDir, line [line.find (' ')+1:])
			elif (line.lower().startswith ('dest-dir')):
				self.destDir = os.path.join (self.baseDir, line [line.find (' ')+1:])
			elif (line.lower().startswith ('ignore-filter')):
				filter = line [line.find (' ')+1:]
				self.log.info ("Adding filter of content to ignore: %s" % filter)
				self.ignoreFilters.append (re.compile (filter))
			elif (line.lower().startswith ('character-set')):
				self.setDefaultCharacterSet (line[line.find (' ') + 1:])
			elif (line.lower().startswith ('additional-plugins-dir')):
				self.extraPluginDirs.append (os.path.join (self.baseDir, line [line.find (' ')+1:]))
			else:
				self.log.warn ("SiteConfig Option %s not supported" % line)
	
	def getPluginModules (self, path):
		try:
			dirList = os.listdir (path)
		except:
			return []
			
		self.log.debug ("Adding %s to the Python path." % path)
		sys.path.insert(0, path)
		foundPlugins = []
		for fileName in dirList:
			try:
				if (os.path.isfile (os.path.join (path,fileName))):
					if (fileName == "__init__.py"):
						self.log.debug ("Skipping init file for plugins dir.")
					elif (fileName.endswith ('.py')):
						pluginModuleName = fileName[:-3]
						plugin = __import__ (pluginModuleName, globals(), locals(), pluginModuleName)
						foundPlugins.append (plugin)
						self.log.info ("Loaded PubTal plugin %s" % pluginModuleName)
					else:
						self.log.debug ("Skipping file %s while looking for plugins." % fileName)
				elif (os.path.isdir (os.path.join (path, fileName))):
					pluginModuleName = fileName
					plugin = __import__ (pluginModuleName, globals(), locals(), pluginModuleName)
					foundPlugins.append (plugin)
					self.log.info ("Loaded PubTal plugin %s" % pluginModuleName)
				else:
					self.log.warn ("Found neither file nor dir in plugin directory.")
			except ImportError, e:
				self.log.warn ("Error loading PubTal plugin %s." % fileName)
				self.log.debug ("Exception was: %s" % str (e))
				
		return foundPlugins

class ContentConfig:
	def __init__ (self, siteConfig):
		self.log = logging.getLogger ("PubTal.ContentConfig")
		self.fileTypeMap = {}
		self.fileMap = {}
		self.directoryMap = {}
		self.currentConfigItem = None
		self.siteConfig = siteConfig
		
	def getPage (self, contentPath):
		page = Page (contentPath, self.siteConfig.contentDir, self.siteConfig.destDir, self.siteConfig.templateDir)
		# Look for file type first
		configItem = self.fileTypeMap.get (contentPath [contentPath.rfind ('.')+1:])
		if (configItem is not None):
			configItem.updatePage (page)
		
		# Now work through directory map
		dirConfigItems = []
		head, tail = os.path.split (contentPath)
		while (tail != ''):
			self.log.debug ("Looking for directory configuration at: %s" % head)
			configItem = self.directoryMap.get (head, None)
			if (configItem is not None):
				dirConfigItems.insert (0,configItem)
			if (self.siteConfig.contentDir.startswith (head)):
				# We have reached the top of the content dir, so stop now
				tail = ''
			else:
				head, tail = os.path.split (head)
			
		for confItem in dirConfigItems:
			self.log.debug ("Updating config using %s" % str (confItem))
			confItem.updatePage (page)
		
		# Now look for this specific file
		configItem = self.fileMap.get (contentPath)
		if (configItem is not None):
			configItem.updatePage (page)
			
		return page
		
	def addFileType (self, fileExtension, extensionConfig):
		""" Adds default configuration options for a file type.
		"""
		if (self.fileTypeMap.has_key (fileExtension)):
			self.log.debug ("Configuration for file type %s already exists, merging." % fileExtension)
			# We already have a config.
			existingConfig = self.fileTypeMap [fileExtension]
			# The existing configuration should take precedence - so merge it into the new config
			extensionConfig.updateConfigItem (existingConfig)
		self.fileTypeMap [fileExtension] = extensionConfig
		
	def startDirective (self, directive, options):
		self.currentDirective = directive
		if (directive == 'CONTENT'):
			self.currentConfigItem = PageConfigItem()
			targetPath = os.path.join (self.siteConfig.contentDir, options)
			if (len (options) > 0 and options[0] == '*'):
				# Wild card, this must be a content file type.
				self.log.debug ("Found filetype configuration item.")
				self.fileTypeMap [options[2:]] = self.currentConfigItem
			elif (os.path.isdir (targetPath)):
				# This is a directory directive
				self.log.debug ("Found directory configuration item: %s" % targetPath)
				if (targetPath.endswith (os.sep)):
					targetPath = targetPath [:-1]
				self.directoryMap [targetPath] = self.currentConfigItem
			else:				
				# This is an individual file
				self.log.debug ("Found file configuration item.")
				self.fileMap [targetPath] = self.currentConfigItem
		else:
			self.currentConfigItem = None
			
	def option (self, line):
		if (self.currentConfigItem is not None):
			if (line.lower().startswith ('macro')):
				mnStart = line.find (' ')
				mnEnd = line.find (' ', mnStart+1)
				macroName = line [mnStart+1:mnEnd]
				macro = line [mnEnd + 1:]
				self.currentConfigItem.addMacro (macroName, os.path.join (self.siteConfig.templateDir, macro))
			elif (line.lower().startswith ('header')):
				mnStart = line.find (' ')
				mnEnd = line.find (' ', mnStart+1)
				headerName = line [mnStart+1:mnEnd]
				header = line [mnEnd + 1:]
				self.currentConfigItem.addHeader (headerName, header)
			else:
				firstSpace = line.find (' ')
				name = line [:firstSpace]
				value = line [firstSpace+1:]
				self.currentConfigItem.setOption (name, value)
				
	def endDirective (self, directive):
		self.currentDirective = None
		self.currentConfigItem = None
		
	def configFinished (self):
		# Some debug messages if enabled.
		if (self.log.isEnabledFor (logging.DEBUG)):
			for fileType in self.fileTypeMap.keys():
				self.log.debug ("FileType %s has Configuration: %s" % (fileType, str (self.fileTypeMap [fileType])))
			for dir in self.directoryMap.keys():
				self.log.debug ("Directory %s has Configuration: %s" % (dir, str (self.directoryMap [dir])))
			for file in self.fileMap.keys():
				self.log.debug ("File %s has Configuration: %s" % (file, str (self.fileMap [file])))

class TemplateConfig:
	def __init__ (self, siteConfig):
		self.log = logging.getLogger ("PubTal.TemplateConfig")
		self.fileTypeMap = {}
		self.fileMap = {}
		self.directoryMap = {}
		self.currentConfigItem = None
		self.siteConfig = siteConfig
		
	def getTemplate (self, templateName):
		""" Get a Template object which hold the configuration information for this template.
		
			Note that the templateName is relative to the templateDirectory - not absolute!
		"""
		
		template = Template (templateName, self.siteConfig.templateDir)
		templatePath = template.getTemplatePath()
		
		# Look for file type first
		configMap = self.fileTypeMap.get (template.getTemplateExtension)
		if (configMap is not None):
			template.options.update (configMap)
		
		# Now work through directory map
		dirConfigItems = []
		head, tail = os.path.split (templatePath)
		while (tail != ''):
			self.log.debug ("Looking for directory configuration at: %s" % head)
			configItem = self.directoryMap.get (head, None)
			if (configItem is not None):
				dirConfigItems.insert (0,configItem)
			if (self.siteConfig.templateDir.startswith (head)):
				# We have reached the top of the content dir, so stop now
				tail = ''
			else:
				head, tail = os.path.split (head)
			
		for confItem in dirConfigItems:
			self.log.debug ("Updating config using %s" % str (confItem))
			template.options.update (confItem)
		
		# Now look for this specific file
		configItem = self.fileMap.get (templatePath)
		if (configItem is not None):
			template.options.update (configItem)
			
		return template
		
	def addFileType (self, fileExtension, extensionConfig):
		self.fileTypeMap [fileExtension] = extensionConfig
		
	def startDirective (self, directive, options):
		self.currentDirective = directive
		if (directive == 'TEMPLATE'):
			self.currentConfigItem = {}
			targetPath = os.path.join (self.siteConfig.templateDir, options)
			if (len (options) > 0 and options[0] == '*'):
				# Wild card, this must be a content file type.
				self.log.debug ("Found filetype configuration item.")
				self.fileTypeMap [options[1:]] = self.currentConfigItem
			elif (os.path.isdir (targetPath)):
				# This is a directory directive
				self.log.debug ("Found directory configuration item: %s" % targetPath)
				if (targetPath.endswith (os.sep)):
					targetPath = targetPath [:-1]
				self.directoryMap [targetPath] = self.currentConfigItem
			else:				
				# This is an individual file
				self.log.debug ("Found file configuration item.")
				self.fileMap [targetPath] = self.currentConfigItem
		else:
			self.currentConfigItem = None
			
	def option (self, line):
		if (self.currentConfigItem is not None):
			firstSpace = line.find (' ')
			name = line [:firstSpace]
			value = line [firstSpace+1:]
			self.currentConfigItem [name] = value
				
	def endDirective (self, directive):
		self.currentDirective = None
		self.currentConfigItem = None
		
	def configFinished (self):
		# Some debug messages if enabled.
		if (self.log.isEnabledFor (logging.DEBUG)):
			for fileType in self.fileTypeMap.keys():
				self.log.debug ("FileType %s has Configuration: %s" % (fileType, str (self.fileTypeMap [fileType])))
			for dir in self.directoryMap.keys():
				self.log.debug ("Directory %s has Configuration: %s" % (dir, str (self.directoryMap [dir])))
			for file in self.fileMap.keys():
				self.log.debug ("File %s has Configuration: %s" % (file, str (self.fileMap [file])))


class PageConfigItem:
	def __init__ (self):
		self.macros = {}
		self.headers = {}
		self.options = {}
		
	def updatePage (self, page):
		page.macros.update (self.macros)
		page.headers.update (self.headers)
		page.options.update (self.options)
		
	def updateConfigItem (self, anotherItem):
		""" This method merges in the configuration options held by another
			item.
		"""
		self.macros.update (anotherItem.macros)
		self.headers.update (anotherItem.headers)
		self.options.update (anotherItem.options)
			
	def addMacro (self, name, template):
		self.macros [name] = template
		
	def addHeader (self, name, header):
		self.headers [name] = header
		
	def setOption (self, name, value):
		self.options [name] = value
		
	def __str__ (self):
		desc = ""
		if (len (self.macros) > 0):
			desc += "Macros: %s\n" % str (self.macros)
		if (len (self.headers) > 0):
			desc += "Headers: %s\n" % str (self.headers)
		return desc

class Page:
	def __init__ (self, source, contentDir, destDir, templateDir):
		# Things we need
		self.source = source
		self.macros = {}
		self.options = {'template': 'template.html'}
		self.headers = {}
		self.destDir = destDir
		
		commonRoot = os.path.commonprefix ([contentDir, source])
		self.relativePath = source[len (commonRoot)+1:]
		# relativePath is of the form: [dir/]file
		depth = 0
		head, tail = os.path.split (self.relativePath)
		while (tail != ''):
			depth += 1
			head, tail = os.path.split (head)
		# We go to far, so knock one off!
		self.depth = depth - 1
		
	def getOption (self, name, defaultValue=None):
		return self.options.get (name, defaultValue)
		
	def getBooleanOption (self, name, defaultValue='0'):
		value = self.getOption (name, str (defaultValue))
		lowerValue = value.lower()
		if (lowerValue == 'y' or lowerValue == 'true'):
			return 1
		if (lowerValue == 'n' or lowerValue == 'false'):
			return 0
			
		try:
			val = int (lowerValue)
			return val
		except:
			return 0

	def getSource (self):
		return self.source
		
	def getRelativePath (self):
		return self.relativePath
		
	def getMacros (self):
		return self.macros
		
	def getHeaders (self):
		return self.headers
		
	def getDepthString (self):
		return "../"*self.depth
		
	def getModificationTime (self):
		""" Returns a tuple of the creation and modification date/time. """
		info = os.stat (self.source)
		return info[stat.ST_MTIME]
		
	def __str__ (self):
		return self.source
		
class Template:
	def __init__ (self, templateName, templateDir):
		self.templateName = templateName
		self.templatePath = os.path.join (templateDir, templateName)
		self.options = {'output-type': 'HTML'}
		
	def getTemplateName (self):
		return self.templateName
		
	def getTemplatePath (self):
		return self.templatePath
		
	def getTemplateExtension (self):
		return os.path.splitext (self.templateName)[1][1:]
		
	def getOption (self, name, defaultValue=None):
		return self.options.get (name, defaultValue)
		
	def getBooleanOption (self, name, defaultValue='0'):
		value = self.getOption (name, str (defaultValue))
		lowerValue = value.lower()
		if (lowerValue == 'y' or lowerValue == 'true'):
			return 1
		if (lowerValue == 'n' or lowerValue == 'false'):
			return 0
			
		try:
			val = int (lowerValue)
			return val
		except:
			return 0
