""" Classes to handle publishing of a PubTal site.

	Copyright (c) 2003 Colin Stewart (http://www.owlfish.com/)
	All rights reserved.
		
	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:
	1. Redistributions of source code must retain the above copyright
	   notice, this list of conditions and the following disclaimer.
	2. Redistributions in binary form must reproduce the above copyright
	   notice, this list of conditions and the following disclaimer in the
	   documentation and/or other materials provided with the distribution.
	3. The name of the author may not be used to endorse or promote products
	   derived from this software without specific prior written permission.
	
	THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
	IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
	
	If you make any bug fixes or feature enhancements please let me know!

"""

try:
	import logging
except:
	import InfoLogging as logging
	
import time, codecs, os, os.path

from simpletal import simpleTAL, simpleTALES, simpleTALUtils
import pubtal

import ContentToHTMLConverter

class PagePublisher:
	def __init__ (self, config):
		self.templateCache = simpleTALUtils.TemplateCache()
		self.templateDir = config.getTemplateDir()
		self.destDir = config.getDestinationDir()
		self.characterSet = config.getDefaultCharacterSet()
		self.templateConfig = config.getTemplateConfig()
		self.config = config
		
		supportedContent = config.getSupportedContent()
		self.supportedContent = {}
		
		self.log = logging.getLogger ("PubTal.PagePublisher")
		
		self.log.info ("Looking for supported content types...")
		for contentType in supportedContent.keys():
			klass = supportedContent [contentType]
			obj = klass (self)
			self.supportedContent [contentType] = obj
			
		self.log.info ("Support for content types: %s" % ", ".join (self.supportedContent.keys()))
			
		# Used in the Context
		self.ContextFunctions = ContextFunctions()
		self.pubTalInfo = {'version': pubtal.__version__
							,'url': "http://www.owlfish.com/software/PubTal/"
							,'linkText': """<p title="PubTal is a template driven web site publisher.">Made with <a href="http://www.owlfish.com/software/PubTal/">PubTal</a> %s</p>""" % pubtal.__version__
						}
			
	def getConfig (self):
		return self.config
		
	def getContentPublisher (self, contentType):
		return self.supportedContent.get (contentType, None)
		
	def publish (self, page):
		contentType = page.getOption ('content-type')
		try:
			publisher = self.supportedContent [contentType]
		except:
			self.log.warn ("Unsupported content type: %s" % contentType)
			return
		
		publisher.publish (page)
		
	def expandTemplate (self, template, context, relativeOutputPath, macros):
		""" Expand the given Template object using the context, writing to the
			output path.
			
			Looks up the character-set for each template and macro.
		"""
		absTemplateName = template.getTemplatePath()
		templateCharset = template.getOption ('character-set', self.characterSet)
		suppressXMLDeclaration = template.getOption ('suppress-xmldecl')
		taltemplate = self.templateCache.getTemplate (absTemplateName, inputEncoding=templateCharset)
		outputPath = os.path.join (self.destDir, relativeOutputPath)
		
		# Handle XHTML DOCTYPE
		xmlDoctype = template.getOption ('xml-doctype', None)
		
		self.ContextFunctions.setCurrentPage (relativeOutputPath)
		context.addGlobal ('ispage', self.ContextFunctions.isPage)
		context.addGlobal ('pubtal', self.pubTalInfo)
		
		# Add macros to the context
		macroTemplates = {}
		for macroName in macros.keys():
			macTemplate = self.templateConfig.getTemplate (macros [macroName])
			macroCharSet = macTemplate.getOption ('character-set', self.characterSet)
			
			mTemp = self.templateCache.getTemplate (macros [macroName], inputEncoding=macroCharSet)
			macroTemplates [macroName] = mTemp.macros
		
		context.addGlobal ('macros', macroTemplates)
		dest = self.openOuputFile (outputPath)
		if (isinstance (taltemplate, simpleTAL.XMLTemplate)):
			if (xmlDoctype is not None):
				taltemplate.expand (context, dest, outputEncoding=templateCharset, docType=xmlDoctype, suppressXMLDeclaration=suppressXMLDeclaration)
				dest.close()
				return
		taltemplate.expand (context, dest, outputEncoding=templateCharset)
		dest.close()
		
	def openOuputFile (self, outputPath):
		""" Creates and required directories and opens a file to the destination path.
		
			This provides a common point for PubTal to note all directories it has
			created and files it has written.  This might one day be used to allow
			automatic upload.
		"""
		# Make directories if required.
		destDir = os.path.split (outputPath)[0]
		if (not os.path.exists (destDir)):
			os.makedirs (destDir)
		dest = open (outputPath, 'wb')
		return dest

class ContextFunctions:
	def __init__ (self):
		self.currentPage = None
		self.isPage = simpleTALES.PathFunctionVariable (self.isCurrentPage)
		
	def setCurrentPage (self, page):
		self.currentPage = page.replace (os.sep, '/')
		
	def isCurrentPage (self, otherPage):
		if (self.currentPage == otherPage):
			return 1
		return 0

class ContentPublisher:
	def __init__ (self, pagePublisher):
		self.pagePublisher = pagePublisher
		self.config = self.pagePublisher.config
		self.contentConfig = self.config.getContentConfig()
		self.templateConfig = self.config.getTemplateConfig()
		self.characterSet = self.config.getDefaultCharacterSet()
		self.characterSetCodec = codecs.lookup (self.characterSet)[1]
		self.destDir = self.config.getDestinationDir()
		self.contentDir = self.config.getContentDir()
		
	def readHeadersAndContent (self, page):
		""" This method reads the source file for this page, and then
			returns the headers defined in this file and the raw
			content of the body of the file.
		"""
		sourceFile = open (page.getSource(), 'r')
		readingHeaders = 1
		headers = {}
		pageCharSet = page.getOption ('character-set', None)
		if (pageCharSet is not None):
			# This page has it's own character set
			pageCodec = codecs.lookup (self.pageCharSet)
		else:
			# This page uses the default character set.
			pageCodec = self.characterSetCodec
		while (readingHeaders):
			line = pageCodec (sourceFile.readline())[0]
			offSet = line.find (':')
			if (offSet > 0):
				headers [line[0:offSet]] = line[offSet + 1:].strip()
			else:
				readingHeaders = 0
		rawContent = pageCodec (sourceFile.read())[0]
		sourceFile.close()
		return (headers, rawContent)
						
	def getPageContext (self, page, template):
		""" Returns the default context which will apply to most pages of
			content.  Template is the template that this context will 
			eventually be used in, and is used to extract the type
			of output (HTML, XHTML, WML, etc) and the destination
			file extension.
		"""
		lastModDate = time.strftime ('%a, %d %b %Y %H:%M:%S %Z', time.localtime (page.getModificationTime()))
		copyrightYear = time.strftime ('%Y')
		
		destExtension = '.' + template.getTemplateExtension()
		relativeDestPath = os.path.splitext (page.getRelativePath())[0] + destExtension
		destPath = os.path.join (self.destDir, relativeDestPath)
		destFilename = os.path.basename (destPath)
		
		pageContext = {'lastModifiedDate': lastModDate
					,'copyrightYear': copyrightYear
					,'sourcePath': page.getRelativePath()
					,'absoluteSourcePath': page.getSource()
					,'destinationPath': relativeDestPath
					,'absoluteDestinationPath': destPath
					,'destinationFilename': destFilename
					,'depth': page.getDepthString()
					,'headers': page.getHeaders()
					}
					
		return pageContext
						
class PublisherException (Exception):
	pass